#!/usr/bin/env python3
"""
    Resources
    =========

    The following constitutes the base-logic of CIJOE:

    * cijoe.core.command (Cijoe)
    * cijoe.core.transport (Transport, Local, SSH)
    * cijoe.core.misc (As the name suggests; various helper-functions)
    * cijoe.cli (Command-Line Tool and utilization of the above for workflow execution)

    Everything else, literally everything, is implemented as dynamically
    collectable and loadable resources. That is, cijoe
    configuration-files/scripts/workflows/templates, and auxiliary files.

    The base-representation of these resources are the cijoe.core.resources.Resource
    class, with content-specific subclasses (Config, Script, and Workflow).

    These resources are collected from installed and locally available Packages, as well
    as for path by the cijoe.core.resources.Collector.

    The Collector is a SingleTon, since it is used extensively everywhere and the tasks
    of doing collection can be somewhat time-consuming.

    Intended usage
    --------------

    From Python::

        from cijoe.core.resources import get_resources
        resources = get_resources()

    From command-line::

        cijoe --resources
"""
import ast
import importlib
import inspect
import logging as log
import os
import pkgutil
import re
import sys
from importlib.machinery import SourceFileLoader
from pathlib import Path
from typing import Any, Dict, List, Union

try:
    from importlib.resources import files as importlib_files
except ImportError:
    from importlib_resources import files as importlib_files_fallback

    importlib_files = importlib_files_fallback

import jinja2
import yaml

import cijoe

if sys.version_info >= (3, 11):
    import tomllib  # Python 3.11 and newer
else:
    import tomli as tomllib  # For older Python versions

import tomli_w


def dict_from_yamlfile(path: Path):
    """Returns content of yamlfile at 'path' as dict and {} on empty document."""

    with path.open("rb") as yamlfile:
        return yaml.safe_load(yamlfile) or {}


def dict_from_tomlfile(path: Path):
    """Returns content of tomfile at 'path' as dict and {} on empty document."""

    with path.open("rb") as tomlfile:
        return tomllib.load(tomlfile) or {}


def dict_to_tomlfile(data, path: Path):
    """Dumps the given 'data' to the given 'path'"""

    with path.open("wb") as tomlfile:
        return tomli_w.dump(data, tomlfile)


def default_context(config=None, resources=None):
    """Return a default context for dict-substitution"""

    if resources is None:
        resources = get_resources()

    return {
        "local": {
            "env": os.environ,
        },
        "config": config.options if config else {},
        "resources": resources,
    }


def dict_substitute(topic: dict, context: dict) -> list:
    """Traverse the given 'topic' replacing {{ foo.bar }} entities with ctx. values"""

    errors = []

    jinja_env = jinja2.Environment(undefined=jinja2.StrictUndefined)
    for key, value in topic.items():
        try:
            if isinstance(value, str):
                topic[key] = jinja_env.from_string(value).render(context)
            elif isinstance(value, list) and all(
                isinstance(line, str) for line in value
            ):
                topic[key] = [
                    jinja_env.from_string(line).render(context) for line in value
                ]
            elif isinstance(value, list) and all(
                isinstance(entity, dict) for entity in value
            ):
                for entity in value:
                    errors += dict_substitute(entity, context)
            elif isinstance(value, dict):
                errors += dict_substitute(value, context)
        except jinja2.exceptions.UndefinedError as exc:
            errors.append(f"Substitution-error: {exc}")

    return errors


class Resource(object):
    """Base representation of a Resource"""

    def __init__(self, path: Path, pkg=None):
        self.path = path.resolve()
        self.pkg = pkg

        self.path = path
        self.content = None

        prefix = ".".join(pkg.name.split(".")[1:-1]) + "." if pkg else ""

        self.ident = f"{prefix}{self.path.stem}"

    def __repr__(self):
        return str(self.path)

    def content_from_file(self):
        """Load resource-content from 'self.path'"""

        with self.path.open("r") as resource_file:
            self.content = resource_file.read()


class Config(Resource):
    """
    Encapsulation of a CIJOE config-file, e.g. 'default-config.toml'

    ivar: options: dict of configuration options populated by load() / from_path()
    """

    SUFFIX = ".toml"

    def __init__(self, path: Path, pkg=None):
        super().__init__(path, pkg)

        self.options: Dict[str, Any] = {}

    def load(self):
        """Populates self.options on success. Returns a list of errors otherwise"""

        # config_dict = dict_from_yamlfile(self.path)
        config_dict = dict_from_tomlfile(self.path)

        errors = dict_substitute(config_dict, default_context())
        if errors:
            return errors

        self.options = config_dict
        return []

    @staticmethod
    def from_path(path, pkg=None):
        """Instantiate a Config from path, returning None on error"""

        path = Path(path).resolve()
        if not path.exists():
            return None

        config = Config(path, pkg)
        errors = config.load()
        if errors:
            return None

        return config


class Script(Resource):
    """Script representation and encapsulation"""

    SUFFIX = ".py"
    NAMING_CONVENTION = ["worklet_entry", "script_entry", "main"]

    def __init__(self, path, pkg=None):
        super().__init__(path, pkg)

        self.func = None
        self.mod = None
        self.mod_name = None

    def content_has_script_func(self):
        """Checks whether the resource-content has the script entry-function"""

        try:
            tree = ast.parse(self.content)
        except SyntaxError:
            return False

        for node in [x for x in ast.walk(tree) if isinstance(x, ast.FunctionDef)]:
            if node.name not in Script.NAMING_CONVENTION:
                log.debug(f"skipping; invalid entry-name({node.name})")
                continue

            argnames = [arg.arg for arg in node.args.args]
            if argnames != ["args", "cijoe", "step"]:
                log.debug(f"skipping; invalid argnames({argnames})")
                continue

            return True

        return False

    def load(self):
        """Loads the module and the script-entry function"""

        if self.func:
            return []

        if not self.content:
            self.content_from_file()

        if not self.content_has_script_func():
            return ["Missing script_entry() function in ast"]

        mod_name = str(Path(self.path).resolve().stem)
        mod_path = str(Path(self.path).resolve())

        parent_path = str(Path(self.path).resolve().parent)
        if parent_path not in sys.path:
            sys.path.append(parent_path)

        spec = importlib.util.spec_from_loader(
            mod_name, SourceFileLoader(mod_name, mod_path)
        )
        mod = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(mod)

        # mod = SourceFileLoader("", str(self.path)).load_module()
        for function_name, function in inspect.getmembers(mod, inspect.isfunction):
            if function_name not in Script.NAMING_CONVENTION:
                continue

            self.mod = mod
            self.mod_name = mod_name
            self.func = function
            return []

        return ["Missing script_entry() function in loaded module"]


class Workflow(Resource):
    SUFFIX = ".yaml"
    STATE_FILENAME = "workflow.state"
    STATE: Dict[str, Any] = {
        "doc": "",
        "tag": "",
        "config": {},
        "steps": [],
        "status": {
            "skipped": 0,
            "failed": 0,
            "passed": 0,
            "elapsed": 0.0,
            "started": 0.0,
        },
    }

    def __init__(self, path, pkg=None):
        super().__init__(path, pkg)

        self.state = None
        self.config = None

    def state_dump(self, path):
        """Dump the current workflow-state to yaml-file"""

        with path.open("w+") as state_file:
            yaml.dump(self.state, state_file)

    @staticmethod
    def dict_normalize(topic: dict):
        """Normalize the workflow-dict, transformation of the 'run' shorthand"""

        errors = []

        if "steps" not in topic:
            errors.append("Missing required top-level key: 'steps'")
            return errors

        for step in topic["steps"]:
            if step["name"].endswith(".py"):
                errors.append(
                    "Illegal step name: Suffix .py reserved for Python scripts"
                )
                return errors

            if "run" not in step.keys():
                continue

            step["uses"] = "core.cmdrunner"
            step["with"] = {"commands": step["run"].splitlines()}

            del step["run"]

        return errors

    @staticmethod
    def dict_lint(topic: dict):
        """Returns a list of integrity-errors for the given workflow-dict(topic)"""

        resources = get_resources()

        errors = []

        for top in set(topic.keys()) - set(["doc", "config", "steps"]):
            errors.append(f"Unsupported top-level key: '{top}'")
            return errors
        for top in ["doc", "steps"]:
            if top not in topic:
                errors.append(f"Missing required top-level key: '{top}'")
                return errors

        step_names = []
        duplicate_names = []
        for step in topic["steps"]:
            if step["name"] in step_names:
                duplicate_names.append(step["name"])
            step_names.append(step["name"])

        if len(duplicate_names):
            errors.append(f"Duplicate step-names: {duplicate_names}")
            return errors

        valid = set(["name", "uses", "with"])
        required = set(["name", "uses"])

        for nr, step in enumerate(topic["steps"], 1):
            keys = set(step.keys())

            missing = required - keys
            if missing:
                errors.append(f"Invalid step({nr}); required key(s): {missing}")
                continue

            if not re.match(r"^([a-zA-Z][a-zA-Z0-9\.\-_]*)", step["name"]):
                errors.append(f"Invalid step({nr}); invalid chars in 'name'")
                continue

            unsupported = keys - valid
            if unsupported:
                errors.append(f"Invalid step({nr}); unsupported keys({unsupported})")
                continue

            if step["uses"] not in resources["scripts"]:
                errors.append(
                    f"Invalid step({nr}); unknown resource: script({step['uses']})"
                )
                continue

        return errors

    def load(self, config: Config, extra_steps: list = []):
        """
        Load the workflow-yamlfile, normalize it, lint it, substitute, then construct
        the object properties
        """

        errors: List[Any] = []

        if self.state:
            return errors

        workflow_dict = dict_from_yamlfile(self.path)
        workflow_dict["steps"] += extra_steps

        errors += Workflow.dict_normalize(workflow_dict)
        if errors:
            return errors

        errors += Workflow.dict_lint(workflow_dict)
        if errors:
            return errors

        errors += dict_substitute(workflow_dict, default_context(config))
        if errors:
            return errors

        state = Workflow.STATE.copy()
        state["doc"] = workflow_dict.get("doc")
        state["config"] = workflow_dict.get("config", {})
        state["steps"] = list(state["steps"])
        for nr, step in enumerate(workflow_dict["steps"], 1):
            step["nr"] = nr
            step["status"] = {
                "skipped": 0,
                "passed": 0,
                "failed": 0,
                "elapsed": 0.0,
                "started": 0.0,
            }
            step["id"] = f"{nr:03}_{step['name']}"

            state["steps"].append(step)

        self.state = state

        return errors


class Collector(object):
    """Collects resources from installed packages and the current working directory"""

    RESOURCES = [
        ("configs", Config.SUFFIX),
        ("templates", ".jinja2"),
        ("workflows", Workflow.SUFFIX),
        ("scripts", Script.SUFFIX),
        ("auxiliary", ".*"),
    ]
    IGNORE = ["__init__.py", "__pycache__", "setup.py"]

    def __new__(cls):
        if not hasattr(cls, "instance"):
            cls.instance = super(Collector, cls).__new__(cls)
            cls.instance.reset()
        return cls.instance

    def __process_candidate(self, candidate: Path, category: str, pkg):
        """Inserts the given candidate"""

        resource: Union[Script, Config, Workflow, Resource]

        if category == "scripts":
            resource = Script(candidate, pkg)
            resource.content_from_file()

            if not resource.content_has_script_func():
                category = "auxiliary"
        elif category == "configs":
            resource = Config(candidate, pkg)
        elif category == "workflows":
            resource = Workflow(candidate, pkg)
        else:
            resource = Resource(candidate, pkg)

        self.resources[category][resource.ident] = resource

    def reset(self):
        self.resources = {category: {} for category, _ in Collector.RESOURCES}
        self.is_done = False

    def collect_from_path(self, path=None, max_depth=2):
        """Collects non-packaged scripts from the given 'path'"""

        path = Path(path).resolve() if path else Path.cwd().resolve()

        base = len(str(path).split(os.sep))

        for candidate in list(path.glob("*")) + list(path.glob("*/*")):
            level = len(str(candidate).split(os.sep))
            if max_depth and level > base + max_depth:
                continue

            if candidate.parent.stem == "auxiliary":
                self.__process_candidate(candidate, "auxiliary", None)
                continue

            for category, suffix in Collector.RESOURCES:
                if candidate.name in Collector.IGNORE:
                    continue
                if candidate.suffix != suffix:
                    continue

                self.__process_candidate(candidate, category, None)

    def collect_from_packages(self, path=None, prefix=None):
        """Collect resources from CIJOE packages at the given 'path'"""

        if prefix is None:
            prefix = ""

        for pkg in pkgutil.walk_packages(path, prefix):
            comp = pkg.name.split(".")[1:]  # drop the 'cijoe.' prefix
            if not (
                pkg.ispkg
                and any(cat in comp for cat, _ in Collector.RESOURCES)
                and len(comp) == 2
            ):  # skip non-resource packages
                continue

            _, category = comp
            for candidate in importlib_files(f"{pkg.name}").iterdir():
                if candidate.name in Collector.IGNORE:
                    continue

                self.__process_candidate(candidate, category, pkg)

    def collect(self, paths=[]):
        """Collect from all implemented resource "sources" """

        if self.is_done:
            return

        self.collect_from_packages(cijoe.__path__, "cijoe.")
        self.collect_from_path()

        for path in paths:
            self.collect_from_path(path, 0)

        cwd = Path().cwd() / ".cijoe"
        if cwd.exists():
            self.collect_from_path(cwd)

        home = Path().home() / ".cijoe"
        if home.exists():
            self.collect_from_path(home)

        self.is_done = True


def get_resources(paths=[]):
    """Returns resources collected by Collector"""

    collector = Collector()
    collector.collect(paths)

    return collector.resources



"""
Transfer and install Linux Kernel .deb
======================================

* Remove any files on remote side in directory "/tmp/kdebs"
* Create directory on remote side at "/tmp/kdebs"
* Transfer all .deb files in "step.with.local_kdebs_dir"
  - Store them on the remote side in "/tmp/kdebs"
* Do `dpkg -i` on all .debs in "/tmp/kdebs"

Retargetable: True
------------------

Transfer from local to remote, the config.cijoe.transport.ssh determines the remote.
"""

import errno
import logging as log
from pathlib import Path


def main(args, cijoe, step):
    deb_root = step.get("with", {}).get("local_kdebs_dir", None)
    if not deb_root:
        return errno.EINVAL

    deb_root = Path(deb_root)
    deb_root = (
        deb_root if deb_root.is_absolute() else Path(args.output).parent / deb_root
    )
    remote_kdebs_dir = "/tmp/kdebs"

    log.info(f"deb_root: '{deb_root}'")

    cijoe.run(f"rm  {remote_kdebs_dir}/*.deb || true")
    cijoe.run(f"mkdir -p {remote_kdebs_dir}")

    for path in deb_root.glob("*.deb"):
        cijoe.put(f"{path}", f"{remote_kdebs_dir}/{path.name}")
        err, _ = cijoe.run(f"[ -f {remote_kdebs_dir}/{path.name} ]")
        if err:
            return err

    cijoe.run(f"ls {remote_kdebs_dir} | grep .deb")

    err, _ = cijoe.run(f"dpkg -i {remote_kdebs_dir}/*.deb")
    if err:
        return err

    return 0



import errno
import logging as log
import os
import shutil
import subprocess
from abc import ABC, abstractmethod
from pathlib import Path

import paramiko
from scp import SCPClient

from cijoe.core.misc import ENCODING
from cijoe.core.resources import Config


class Transport(ABC):
    @abstractmethod
    def run(self, cmd, cwd, env: dict, cmd_output):
        pass

    @abstractmethod
    def get(self, src, dst=None):
        pass

    @abstractmethod
    def put(self, src, dst=None):
        pass


class Local(Transport):
    """Provide cmd/push/pull locally"""

    def __init__(self, config: Config, output_path: Path):
        self.config = config
        self.output_path = output_path
        self.output_ident = "artifacts"

    def run(self, cmd, cwd, env, cmd_output):
        """Invoke the given command"""

        env = dict(os.environ)
        env.update(env)

        with subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            shell=True,
            cwd=cwd,
            env=env,
        ) as process:
            while process.poll() is None:
                cmd_output.write(process.stdout.read(1))

            cmd_output.write(process.stdout.read())
            cmd_output.flush()
            process.wait()

            return process.returncode

    def put(self, src, dst=None):
        """..."""

        if dst is None:
            dst = os.path.basename(src)
        if not os.path.isabs(src):
            src = os.path.join(self.output_path, self.output_ident, src)
        if not os.path.isabs(dst):
            dst = os.path.join(self.output_path, self.output_ident, dst)

        if src == dst:
            return True

        if os.path.isdir(src):
            shutil.copytree(src, dst)
            return True

        shutil.copy(src, dst)
        return True

    def get(self, src, dst=None):
        """..."""

        return self.put(src, dst)


class SSH(Transport):
    """Provide cmd/push/pull over SSH"""

    def __init__(self, config, output_path):
        """Initialize the CIJOE SSH Transport"""

        self.config = config
        self.shell = (
            self.config.options.get("cijoe", {}).get("run", {}).get("shell", "sh")
        )
        self.output_path = output_path

        self.ssh = paramiko.SSHClient()

        # Using the 'AutoAddPolicy()' *without* load_system_host_keys(), by doing so,
        # then Paramiko does not know any hosts, and simply adds them first time they
        # are connected to.
        # It was attempted to use load_system_host_keys() with WarningPolicy(), however,
        # when a host changed, e.g. re-provisioned virtual machine, then the host-key
        # changes and Paramiko cannot connect.
        self.ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self.scp = None

        # Not connecting at this point, since the remote side might be ready at the time
        # the transport is initialized. For example, when a qemu-guest is booted, then
        # cijoe.run_local() is used, and not until the guest is up will the cij.run() be
        # used. Also, when a system reboots etc. then dropped connections must be
        # handled, lastly connection close must happen as Python terminates, dangling
        # connections will make it hang.
        # Thus, __connect()/__disconnect() is called for each call to run()/get()/put().
        # Current short-coming is of course that then these cannot happen in parallel.

        log.getLogger("paramiko.transport").setLevel(log.CRITICAL)
        paramiko.util.log_to_file(
            self.output_path / "paramiko.log", level=log.root.level
        )

    def __connect(self):
        self.ssh.connect(
            **self.config.options.get("cijoe", {}).get("transport").get("ssh")
        )
        self.scp = SCPClient(self.ssh.get_transport())

    def __disconnect(self):
        self.scp.close()
        self.ssh.close()

    def run(self, cmd, cwd, env, cmd_output):
        """Invoke the given command"""

        # Add environment variables defined in config.
        env = self.config.options.get("cijoe", {}).get("run", {}).get("env", {}) | env

        # Seems like paramiko 'exec_command' or just SSH Accept-something... does not
        # like setting environment variables... thus.. this injection of them...
        # unfortunately then this is shell-dependent and probably breaks.
        # This is why the CIJOE_DISABLE_SSH_ENV_INJECT is here.
        if self.shell == "csh":
            prefix_env = "".join(
                [f'setenv {key} "{val}"; ' for key, val in env.items()]
            )
        elif self.shell == "cmd":
            prefix_env = "".join([f"set {key}={val} && " for key, val in env.items()])
        elif self.shell == "pwsh":
            prefix_env = "".join(
                [f"$env:{key} = '{val}'; " for key, val in env.items()]
            )
        else:
            prefix_env = "".join([f'{key}="{val}" ' for key, val in env.items()])
        if os.environ.get("CIJOE_DISABLE_SSH_ENV_INJECT", None):
            prefix_env = ""

        cmd = f"{prefix_env}{cmd}"
        if cwd:
            cmd = f"cd {cwd}; {cmd}"

        try:
            self.__connect()

            _, stdout, _ = self.ssh.exec_command(cmd, environment=env)
            stdout.channel.set_combine_stderr(True)

            while not stdout.channel.exit_status_ready():
                cmd_output.write(stdout.read(1))

            cmd_output.write(stdout.read())
            cmd_output.flush()

            err = stdout.channel.recv_exit_status()

            self.__disconnect()
        except paramiko.ssh_exception.SSHException as exc:
            err = (
                exc.errno
                if hasattr(exc, "errno")
                and isinstance(exc.errno, int)
                and exc.errno > 0
                else errno.EIO
            )
            log.error(f"ssh-err({exc})")
            log.debug(f"cmd({cmd}), env({env})")
            log.debug(
                f"run():{type(exc).__name__}, {__file__}:{exc.__traceback__.tb_lineno}"
            )

        return err

    def put(self, src, dst=None):
        """Hmm... no return-value just exceptions"""

        self.__connect()

        if dst is None:
            dst = os.path.basename(src)
        if not os.path.isabs(src):
            src = os.path.join(self.output_path, self.output_ident, src)

        self.scp.put(src, dst, recursive=True)

        self.__disconnect()

        return True

    def get(self, src, dst=None):
        """Hmm... no return-value just exceptions"""

        self.__connect()

        if dst is None:
            dst = os.path.basename(src)
        if not os.path.isabs(src):
            dst = os.path.join(self.output_path, self.output_ident, dst)

        self.scp.get(src, dst, recursive=True)

        self.__disconnect()

        return True



"""
insert / remove null_blk
========================

Insert or remove null_blk instances, based on the value of step.args.do

* steps.args.do == "insert"
  - Insert the nullblk module

* step.args.do == "remove"
  - Remove the nullblk module

Retargetable: True
------------------
"""

import errno

import cijoe.linux.null_blk as null_blk


def main(args, cijoe, step):
    """Insert or remove the null_blk"""

    do = step.get("with", {"do": "insert"}).get("do", "insert")
    if do == "insert":
        err, _ = null_blk.insert(cijoe)
    elif do == "remove":
        err, _ = null_blk.remove(cijoe)
    else:
        err = errno.EINVAL

    return err



"""

"""

import errno
import logging as log
import os
import sys
import time
import traceback
from pathlib import Path
from typing import Any, Union

import yaml

from cijoe.core import transport
from cijoe.core.misc import ENCODING, sanitize_ident
from cijoe.core.resources import Config


def convert_str(value: str) -> Union[str, int, bool]:
    # Strip and lowercase the input to make checks more robust
    lower_value = value.lower().strip()

    # Check for boolean values
    if lower_value in ("true", "false"):
        return lower_value == "true"

    # Check for hexadecimal numbers (starting with '0x' or '0X')
    if lower_value.startswith("0x"):
        try:
            return int(value, 16)  # Convert hex to int
        except ValueError:
            pass  # In case the hex is invalid, we just fall through

    # Check for integer conversion (decimal numbers)
    if value.isdigit() or (value.startswith("-") and value[1:].isdigit()):
        return int(value)

    # If no conversion is possible, return the original string
    return value


def default_output_path():
    """Returns a default output-path"""

    return os.path.join(os.getcwd(), "cijoe-output")


class Tee:
    def __init__(self, file_path, monitor):
        self.file = open(file_path, "w")
        self.monitor = monitor
        self.buffer = []
        self.buffer_size = 0

    def write(self, data: bytes):
        decoded = data.decode(ENCODING, errors="replace")
        self.buffer.append(decoded)
        self.buffer_size += len(decoded)

        if "\n" in decoded or self.buffer_size >= 4096:
            self.flush()

    def flush(self):
        output = "".join(self.buffer)

        self.file.write(output)
        self.file.flush()

        if self.monitor:
            sys.stdout.write(output)
            sys.stdout.flush()

        self.buffer = []
        self.buffer_size = 0

    def close(self):
        self.file.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


class CommandState(object):
    def __init__(self, cijoe, cmd, cwd, err, begin, end, is_done, monitor):
        self.cmd = cmd
        self.cwd = cwd
        self.err = err
        self.begin = begin
        self.end = end
        self.elapsed = end - begin
        self.is_done = is_done

        cmd_output_dpath = os.path.join(cijoe.output_path, cijoe.output_ident)
        cmd_output_fpath = os.path.join(
            cmd_output_dpath, f"cmd_{cijoe.run_count:02}.output"
        )
        cmd_state_fpath = os.path.join(
            cmd_output_dpath, f"cmd_{cijoe.run_count:02}.state"
        )
        os.makedirs(cmd_output_dpath, exist_ok=True)

        self.output_dpath = Path(cmd_output_dpath)
        self.output_fpath = Path(cmd_output_fpath)
        self.state_fpath = Path(cmd_state_fpath)

        self.cmd_output = Tee(self.output_fpath, monitor)

    def output(self):
        """Returns the content of 'output_fpath'"""

        with self.output_fpath.open() as ofd:
            return ofd.read()

    def to_file(self):
        """Dump the command state to file"""

        with self.state_fpath.open("w", encoding=ENCODING) as state_file:
            state = {
                "cmd": self.cmd,
                "cwd": str(self.cwd),
                "err": self.err,
                "begin": self.begin,
                "end": self.end,
                "elapsed": self.elapsed,
                "output_dpath": str(self.output_dpath),
                "output_fpath": str(self.output_fpath),
                "is_done": self.is_done,
            }
            yaml.dump(state, state_file)


class Cijoe(object):
    """CIJOE providing retargetable command-line expressions and data-transfers"""

    def __init__(self, config: Config, output_path: Path, monitor: bool):
        """Create a cijoe encapsulation defined by the given config_fpath"""

        self.config = config

        self.run_count = 0
        self.output_path = output_path if output_path else default_output_path()
        self.output_ident = "artifacts"

        self.monitor = monitor

        os.makedirs(os.path.join(self.output_path, self.output_ident), exist_ok=True)

        self.transport_local = transport.Local(self.config, self.output_path)
        self.transport = self.transport_local

        ssh = self.config.options.get("cijoe", {}).get("transport", {}).get("ssh", None)
        if ssh:
            self.transport = transport.SSH(self.config, self.output_path)

    def set_output_ident(self, output_ident: str):
        """
        This sets the output-identifier which is used in order to provide a subfolder
        for artifacts, command-output etc. Additionally, then it reset the command
        run-count
        """

        output_ident = sanitize_ident(output_ident)

        self.run_count = 0
        self.output_ident = output_ident
        self.transport.output_ident = output_ident

    def _run(self, cmd, cwd, env, transport):
        self.run_count += 1

        begin = time.time()
        state = CommandState(
            cijoe=self,
            cmd=cmd,
            cwd=cwd,
            err=0,
            begin=begin,
            end=0,
            is_done=False,
            monitor=self.monitor,
        )
        state.to_file()

        with state.cmd_output as cmd_output:
            err = transport.run(cmd, cwd, env, cmd_output)
            state.err = err
            state.end = time.time()
            state.elapsed = state.end - state.begin
            state.is_done = True
            state.to_file()

        return err, state

    def run(self, cmd, cwd=None, env={}):
        """
        Execute the given shell command/expression via 'config.transport'

        Commands executed using this will write stdout and stderr to file. The location
        of the cmd_output is fixed to: "output_path/output_ident/cmd_XX.output", such that the
        location is a subfolder of the output_path. Unless somebody wants to break the
        convention and call set_output_ident("../..")
        """

        return self._run(cmd, cwd, env, self.transport)

    def run_local(self, cmd, cwd=None, env={}):
        """
        Execute the given shell command/expression via local transport

        Commands executed using this will write stdout and stderr to file. The location
        of the cmd_output is fixed to: "output_path/output_ident/cmd_XX.output", such that the
        location is a subfolder of the output_path. Unless somebody wants to break the
        convention and call set_output_ident("../..")
        """

        return self._run(cmd, cwd, env, self.transport_local)

    def put(self, src, dst):
        """Transfer 'src' on 'dev_box' to 'dst' on **test_target**"""

        os.makedirs(os.path.join(self.output_path, self.output_ident), exist_ok=True)

        try:
            return self.transport.put(src, dst)
        except Exception as exc:
            log.error(f"err({exc})")
            log.debug(f"src({src}), dst({dst})")
            log.debug(
                f"run():{type(exc).__name__}, {__file__}:{exc.__traceback__.tb_lineno}"
            )

        return False

    def get(self, src, dst):
        """Transfer 'src' on 'test_target' to 'dst' on **dev_box**"""

        os.makedirs(os.path.join(self.output_path, self.output_ident), exist_ok=True)

        try:
            return self.transport.get(src, dst)
        except Exception as exc:
            log.error(f"err({exc})")
            log.debug(f"src({src}), dst({dst})")
            log.debug(
                f"run():{type(exc).__name__}, {__file__}:{exc.__traceback__.tb_lineno}"
            )

        return False

    def getconf(self, key: str, default: Any = None):
        """
        Return value for given key, and return default if no value is found.

        The key must be a sequence of namespaces separated by a point, ex.
        "foo.bar.jazz".

        The value is found in the cijoe configuration file, but is overwritten
        if the key (with points replaced with underscores, ex. FOO_BAR_JAZZ) is
        found in the initiator's environment variables.
        """

        envkey = key.replace(".", "_").upper()
        envvar = os.getenv(envkey)
        if envvar:
            log.debug(f"found {key} ({envkey}) in environment variables.")
            return convert_str(envvar)

        dict_keys = key.split(".")
        config = self.config.options

        try:
            # fold over the list of dict keys, to iterate through nested dicts
            [config := config.get(key, {}) for key in dict_keys[:-1]]
            return config.get(dict_keys[-1], default)
        except AttributeError as exc:
            # AttributeError is raised if one of the keys are associated to a
            # value that is not a dict.
            log.debug(f"{key} not in config: {exc}")
            return default



"""
Download and unpack the GitHub Action Runner

* This script will download GitHub Action Runner from 'gha.runner.url'
* Unpack the GitHub Action to 'gha.runner.home / ghar' associated

Below is an example of the configuration items that this script uses::

    [gha.runner]
    url = "https://github.com/actions/runner/releases/download/v2.319.1/actions-runner-linux-x64-2.319.1.tar.gz"
    home = "/opt/gha"
"""

import logging as log
import sys
from pathlib import Path

import requests


def url_to_latest_runner_on_x86():
    tag_name = requests.get(
        "https://api.github.com/repos/actions/runner/releases/latest"
    ).json()["tag_name"]

    return (
        f"https://github.com/actions/runner/releases/download/{tag_name}/"
        f"actions-runner-linux-x64-{tag_name[1:]}.tar.gz"
    )


def main(args, cijoe, step):
    runner = cijoe.config.options.get("gha", {}).get("runner", {})

    url = runner.get("url", {}).get("download", url_to_latest_runner_on_x86())
    home = runner.get("home", None)
    if None in [url, home]:
        log.error(f"missing or invalid config gha.runner({runner})")
        return 1

    for cmd in [
        f"mkdir -p {home}/ghar",
        f"mkdir -p {home}/runners",
        f"curl -o {home}/ghar.tar.gz -L {url}",
        f"tar xzf {home}/ghar.tar.gz -C {home}/ghar",
    ]:
        err, _ = cijoe.run(cmd)
        if err:
            log.error(f"cmd({cmd}), err({err})")
            return err

    return 0



"""
    kmemleak module, providing helpers to: "clear", "scan" and "cat" kmemleak

    For details on kmemleak, have a look at the Linux Kernel Documentation at

    https://www.kernel.org/doc/html/latest/dev-tools/kmemleak.html

    retargetable: True
"""


def cat(cijoe):
    """Dump the contents of kmemleak"""

    return cijoe.run("cat /sys/kernel/debug/kmemleak")


def scan(cijoe):
    """Scan the kernel"""

    return cijoe.run("echo scan > /sys/kernel/debug/kmemleak")


def clear(cijoe):
    """Clear the kmemleak"""

    return cijoe.run("echo clear > /sys/kernel/debug/kmemleak")



"""
get
===

Copies a file from remote to local.

Step Arguments
--------------

step.with.src: path to the file on remote machine
step.with.dst: path to where the file should be placed on the local machine

Retargetable: True
------------------
"""

import errno


def main(args, cijoe, step):
    """Copies the file at step.with.src on the remote machine to step.with.dst on the local machine"""

    if not ("with" in step and "src" in step["with"] and "dst" in step["with"]):
        return errno.EINVAL

    return int(not cijoe.get(step["with"]["src"], step["with"]["dst"]))



#!/usr/bin/env python3
"""
build qemu-system(x86_64 aarch64)
=================================

In the build, virtfs and debugging enabled.

Arguments
---------

* repository.path
* build.prefix

Retargetable: False
-------------------
"""
import errno
import logging as log
from pathlib import Path


def main(args, cijoe, step):
    """Build qemu"""

    conf = cijoe.config.options.get("qemu", None)
    if not conf:
        log.error("config is missing 'qemu' section")
        return errno.EINVAL

    build_dir = Path(conf["repository"]["path"]) / "build"

    configure_args = [
        f"--prefix={conf['build']['prefix']}",
        "--audio-drv-list=''",
        "--disable-docs",
        "--disable-glusterfs",
        "--disable-libnfs",
        "--disable-libusb",
        "--disable-opengl",
        "--disable-sdl",
        "--disable-smartcard",
        "--disable-spice",
        "--disable-virglrenderer",
        "--disable-vnc",
        "--disable-vte",
        "--disable-xen",
        "--enable-debug",
        "--enable-virtfs",
        "--target-list=x86_64-softmmu,aarch64-softmmu",
    ]

    err, _ = cijoe.run_local(f"mkdir -p {build_dir}")
    if err:
        return err

    err, _ = cijoe.run_local("../configure " + " ".join(configure_args), cwd=build_dir)
    if err:
        return err

    err, _ = cijoe.run_local("make -j $(nproc)", cwd=build_dir)
    if err:
        return err

    return 0



#!/usr/bin/env python3
"""
install fio
===========

Just a plain 'make install' within 'repository.path'

Retargetable: True
------------------
"""
from pathlib import Path


def main(args, cijoe, step):
    """Install fio"""

    err, _ = cijoe.run(
        "make install", cwd=Path(cijoe.config.options["fio"]["repository"]["path"])
    )
    return err



"""
Errors that are used across cijoe modules.

NOTE: new errors should inherit from CIJError. This allows users to catch all
cijoe-related errors using a single type.
"""


class CIJError(Exception):
    """Wrapper for all CIJOE errors"""


class InitializationError(CIJError):
    """Raised when failing to initialize data structures"""


class InvalidRangeError(CIJError):
    """Raised for range formatting errors"""


class UnknownUnitError(CIJError):
    """Raised when attempting to use unknown units"""



"""
put
===

Copies a file from local to remote.

Step Arguments
--------------

step.with.src: path to the file on local machine
step.with.dst: path to where the file should be placed on the remote machine

Retargetable: True
------------------
"""

import errno


def main(args, cijoe, step):
    """Copies the file at step.with.src on the local machine to step.with.dst on the remote machine"""

    if not ("with" in step and "src" in step["with"] and "dst" in step["with"]):
        return errno.EINVAL

    return int(not cijoe.put(step["with"]["src"], step["with"]["dst"]))



"""
testrunner
==========

This is the CIJOE testrunner, it is implemented as a pytest-plugin. By doing
so, then there is little to nothing to learn, assuming, that you have used
pytest before.

The "special" thing is the pytest-plugin, it provides pytest command-line
arguments: ``-c / --config`` and ``-o / --output``. These behave just like the
``cijoe`` counter-part. Within, your tests, then a ``cijoe`` fixture is
provided. It will give the tests access to a cijoe-instance with the given
config. Quite neat :)

Thus, arbitrary testing of third-party test-suites, tools benchmarks etc. is
possible with something as simple as::

    def test_foo(cijoe):

        err, state = cijoe.run("execute-some-testsuite")
        assert not err

Requires the following pytest plugins for correct behaviour:

* cijoe, fixtures providing 'cijoe' object and "--config" and "--output"
  pytest-arguments to instantiate cijoe.

* report-log, dump testnode-status as JSON, this is consumed by 'core.report'
  to produce an overview of testcases and link them with the cijoe-captured
  output and auxiliary files.

Invocation of pytest is done in one of the following two ways, and controlled
by ``step.with.run_local``, with boolean value True / False. The default is
True.

with.run_local: True
--------------------

This is the most common, invoking pytest locally, which in turn will be using
the same config as all other cijoe-scripts. To clarify, cijoe will execute
'pytest' in the same environment/system where the ``cijoe`` cli was executed.

with.run_local: False
---------------------

This is a special-case, where a collection of pytests uses cijoe, but only the
configuration, the that the pytest verify is Python code / statements /
expressions, not CIJOE command executions cijoe.run(). In order ot run these
remotely, then the code must be available, and then it does the following:

* Create a copy of the currently used cijoe-config
  - Remove the transport section if any is there
* Transfer cijoe-config-copy to remote
* Invoke pytest remotely using cijoe-config-copy
* Download the testreport.log from remote

Why this? This allows for executing pytests on a remote system which does not
use cijoe.run(). Such as tests implemented in Python.
"""

import copy
import logging as log
import uuid
from pathlib import Path

from cijoe.core.resources import dict_to_tomlfile


def pytest_cmdline(args, step, config_path, output_path, reportlog_path):
    """Contruct pytest command-line arguments given 'step' and paths"""

    log.info(f"config_path({config_path})")
    log.info(f"output_path({output_path})")
    log.info(f"reportlog_path({reportlog_path})")

    cmdline = ["pytest"]
    if args.config:
        cmdline.append("--config")
        cmdline.append(config_path)
    cmdline += ["--output", output_path]
    cmdline += ["--report-log", reportlog_path]

    cmdline += step.get("with").get("args", "").split(" ")

    cmdline += ["--random-order"] if step.get("with").get("random-order", True) else []

    return cmdline


def pytest_remote(args, cijoe, step):
    """
    Run pytest on remote, that is, transfer config, execute pytest on the
    remote and transfer the testreport.log from remote to local
    """

    err, state = cijoe.run("pwd")
    if err:
        log.error("Failed querying 'pwd'")
        return err

    cwd = Path(state.output().strip())
    rand = str(uuid.uuid4())[:8]

    # Construct config based on current config, but without "ssh" transport
    config_stem = f"cijoe-config-{rand}.toml"
    config_path = cwd / config_stem
    config_path_local = args.output / cijoe.output_ident / config_stem

    config = copy.deepcopy(cijoe.config.options)
    if "transport" in config.get("cijoe", {}) and "ssh" in config["cijoe"]["transport"]:
        del config["cijoe"]["transport"]["ssh"]

    dict_to_tomlfile(config, config_path_local)
    cijoe.put(str(config_path_local), config_path)

    # Construct pytest command-line and execute it
    reportlog_stem = f"cijoe-testreport-{rand}.log"
    reportlog_path = cwd / reportlog_stem
    reportlog_path_local = args.output / cijoe.output_ident / "testreport.log"

    # Construct the output-path and the equivalent local-path
    output_stem = f"cijoe-output-{rand}"
    output_path = cwd / output_stem
    output_path_local = args.output / cijoe.output_ident / "output_pytest"

    cmdline = pytest_cmdline(
        args,
        step,
        str(config_path),
        str(output_path),
        str(reportlog_path),
    )

    err, _ = cijoe.run(" ".join(cmdline))

    # Retrieve testlog
    cijoe.get(str(reportlog_path), str(reportlog_path_local))
    log.info(reportlog_path_local)

    # TODO: Retrieve output
    cijoe.get(str(output_path), str(output_path_local))

    # Cleanup: remove "artifacts" on remote
    cijoe.run(f"rm {config_path}")
    cijoe.run(f"rm {reportlog_path}")
    cijoe.run(f"rm -r {output_path}")

    return err


def pytest_local(args, cijoe, step):
    """
    Run pytest locally, that is, execute on the same system on which cijoe-cli
    was executed. The cijoe config provided via 'args.config' is forwarded to
    the pytest-cijoe-plugin unmodified.
    """

    cmdline = pytest_cmdline(
        args,
        step,
        str(args.config),
        str(args.output / cijoe.output_ident),
        str(args.output / cijoe.output_ident / "testreport.log"),
    )

    err, _ = cijoe.run_local(" ".join(cmdline))

    return err


def main(args, cijoe, step):
    """Invoke the pytest + cijoe-plugin test-runner"""

    return (
        pytest_local
        if step.get("with", {"run_local": True}).get("run_local", True)
        else pytest_remote
    )(args, cijoe, step)



#!/usr/bin/env python3
"""
Install qemu
============

Retargetable: False
-------------------
"""
import errno
from pathlib import Path


def main(args, cijoe, step):
    """Install qemu"""

    conf = cijoe.config.options.get("qemu", None)
    if not conf:
        return errno.EINVAL

    build_dir = Path(conf["repository"]["path"]) / "build"

    err, _ = cijoe.run_local("make install", cwd=build_dir)

    return err



#!/usr/bin/env python3
"""
Create ``boot.img`` from a cloud-init image
===========================================

Initialize the guest's boot image (``boot.img``) using a
cloud-init image. When the cloud-init image at
``qemu.guest.init_using_cloudinit.img`` does not
exist, then it is downloaded from
``qemu.guest.init_using_cloudinit.url`` and stored at
``qemu.guest.init_using_cloudinit.img`` for re-use.

Config::

    [qemu.guest.init_using_cloudinit]
    url = # URL of cloud-init image, e.g. on https://cloud.debian.org/images/cloud/
    img = # Path to cloud-init image
    meta = # Path to cloud-init meta-file
    user = # Path to cloud-init user-file
    pubkey = # SSH public-key to inject into guest, (when provided)

Retargetable: False
-------------------
"""
from cijoe.qemu.wrapper import Guest


def main(args, cijoe, step):
    """Provision a qemu-guest using a cloud-init image"""

    guest = Guest(cijoe, cijoe.config)

    return guest.init_using_cloudinit()



"""
Linux Custom Kernel as Debian Package
=====================================

There are a myriad of ways to build and install a custom Linux kernel. This worklet
builds it as a Debian package. The generated .deb packages are stored in
cijoe.output_path.

Retargetable: False
-------------------

It is intended to be run "locally" since, currently the collection of the generated
.debs are not retrieved via cijoe.get(), doing so would make it retargetable.

Worklet arguments
-----------------

with.localversion
"""

from pathlib import Path


def main(args, cijoe, step):
    """Configure, build and collect the build-artifacts"""

    repos = Path(cijoe.config.options["linux"]["repository"]["path"]).resolve()
    err, _ = cijoe.run(f"[ -d {repos} ]")
    if err:
        return err

    localversion = step.get("with", {}).get("localversion", "custom")
    run_local = step.get("with", {}).get("run_local", True)
    run = cijoe.run_local if run_local else cijoe.run

    commands = [
        "[ -f .config ] && rm .config || true",
        "yes " " | make olddefconfig",
        "./scripts/config --disable CONFIG_DEBUG_INFO",
        "./scripts/config --disable SYSTEM_TRUSTED_KEYS",
        "./scripts/config --disable SYSTEM_REVOCATION_KEYS",
        f"yes '' | make -j$(nproc) bindeb-pkg LOCALVERSION={localversion}",
        f"mkdir -p {cijoe.output_path}/artifacts/linux",
        f"mv ../*.deb {cijoe.output_path}/artifacts/linux",
        f"mv ../*.changes {cijoe.output_path}/artifacts/linux",
        f"mv ../*.buildinfo {cijoe.output_path}/artifacts/linux",
    ]
    for cmd in commands:
        err, _ = run(cmd, cwd=str(repos))
        if err:
            return err

    return 0



from pathlib import Path

import cijoe.core
from cijoe.core.resources import Collector

CORE_RESOURCE_COUNTS = {
    "configs": 2,
    "templates": 2,
    "auxiliary": 2,
    "scripts": 7,
}


def test_resource_collection():
    """Check that the expected amount of resources are collected"""

    collector = Collector()
    collector.collect()

    for category, count in CORE_RESOURCE_COUNTS.items():
        assert (
            len(
                [
                    r
                    for r in collector.resources[category].keys()
                    if r.startswith("core")
                ]
            )
            == count
        ), f"Invalid count({count}) for category({category})"


def test_collect_scripts_from_path():
    """Uses the core package, to have something to collect."""

    collector = Collector()
    collector.collect_from_path(Path(__file__).parent.parent.joinpath("scripts"))

    assert (
        len(collector.resources["scripts"]) == CORE_RESOURCE_COUNTS["scripts"]
    ), "Failed collecting from path"


def test_collect_scripts_from_packages():
    collector = Collector()
    collector.collect_from_packages(cijoe.core.__path__, cijoe.core.__name__ + ".")

    assert (
        len(collector.resources["scripts"]) == CORE_RESOURCE_COUNTS["scripts"]
    ), "Failed collecting from packages"


def test_compare_from_path_with_from_package():
    """This is just to give hint to whether it is 'from_path' or 'from_packages'"""

    collector_path = Collector()
    collector_path.collect_from_path(Path(__file__).parent.parent.joinpath("scripts"))

    collector_pkgs = Collector()
    collector_pkgs.collect_from_packages(cijoe.core.__path__, cijoe.core.__name__ + ".")

    assert len(collector_path.resources["scripts"]) == len(
        collector_pkgs.resources["scripts"]
    )


def test_collect_from_empty_path():
    """This should return an empty dictionary"""

    collector = Collector()
    collector.collect_from_path("/tmp")

    assert len(collector.resources["scripts"]) == 0, "Did not expect to find any"



"""
Unregister a batch of self-hosted GitHub Action Runners

* This will stop runner-services, uninstall and remove them

Below is an example of the configuration items that this script uses::

    [gha.runner]
    home = "/opt/gha"
    nameprefix = "qemu-host"
    count = 10

Additionally, then ensure that the following environment variables are set::

    GHA_RUNNER_TOKEN
"""

import logging as log
import os
import sys
from pathlib import Path


def main(args, cijoe, step):
    runner = cijoe.config.options.get("gha", {}).get("runner", {})

    home = runner.get("home", None)
    count = runner.get("count", None)
    nameprefix = runner.get("nameprefix", None)
    token = os.getenv("GHA_RUNNER_TOKEN", runner.get("token", None))

    if None in [home, count, nameprefix, token]:
        log.error(f"missing or invalid config gha.runner({runner})")
        return 1

    errors = []
    for number in range(int(count)):
        name = f"{nameprefix}{number:02d}"
        rdir = f"{home}/runners/{name}"

        for cmd in [
            "sudo ./svc.sh stop",
            "sudo ./svc.sh uninstall",
            f"./config.sh remove --token {token}",
        ]:
            err, _ = cijoe.run(cmd, cwd=rdir)
            if err:
                log.error(f"cmd({cmd}), err({err})")
                errors.append(err)

    return errors[0] if errors else 0



#!/usr/bin/env python3
"""
check fio
=========

Check the version of the wrapped fio. This is useful for reference and as a check that
fio-wrapper has the correct options set.

Retargetable: True
------------------
"""
from cijoe.fio.wrapper import fio


def main(args, cijoe, step):
    """Check version of fio"""

    err, _ = fio(cijoe, "--help")

    return err



"""
collect Linux system information
================================

Collects a bunch of information about the system kernel and hardware.

Retargetable: True
------------------
"""


def main(args, cijoe, step):
    """Collect Linux system information"""

    commands = [
        "hostname",
        "lsb_release --all || cat /etc/os-release",
        "uname -a",
        "cat /boot/config-$(uname -r)",
        "set",
        "lsblk",
        "lscpu",
        "lslocks",
        "lslogins",
        "lsmem",
        "lsmod",
        "lspci",
    ]

    err = 0
    for cmd in commands:
        err, state = cijoe.run(cmd)
        if err:
            err = err

    return err



from cijoe.core.resources import get_resources


def test_workflow_load():
    resources = get_resources()

    config = resources["configs"]["core.default-config"]
    assert config

    errors = config.load()
    assert not errors

    workflow = resources["workflows"]["core.example-workflow"]
    assert workflow

    errors = workflow.load(config, [])
    assert not errors



"""
cijoe example script
====================

An example of using the core infrastructure of cijoe:

* cijoe.run(command)

  - Error-handling; checking return-code
  - Output processing state.output()

cijoe also has primitives for transferring data:

* cijoe.get(src, dst)

  - Transfer 'src' directory or file from **target** to 'dst' on **initiator** 

* cijoe.put(src, dst)

  - Transfer 'src' directory or file from **initiator** to 'dst' on **target** 

These are not used in the example code below, but you can experiment and try adding
them yourself.
"""

import logging as log
from argparse import Namespace

from cijoe.core.command import Cijoe


def main(args: Namespace, cijoe: Cijoe, step: dict):
    """Entry-point of the cijoe-script"""

    err, state = cijoe.run("echo 'Hello World!'")
    if "Hello" not in state.output():
        log.error("Something went wrong")
        return 1

    log.info(f"All is good? err({err})")

    return err



import pytest

from cijoe.core.resources import Collector


@pytest.fixture
def nvme(cijoe, capsys):
    nvme = cijoe.get_config(subject="nvme")
    if not nvme:
        return None

    for device in nvme.get("devices", []):
        return device


@pytest.fixture
def xnvme(cijoe, capsys):
    nvme = cijoe.get_config(subject="nvme")
    if not nvme:
        return None

    for device in nvme.get("devices", []):
        return device


@pytest.fixture(autouse=True)
def reset_collector():
    # This fixture runs automatically before each test
    collector = Collector()
    collector.reset()
